/***************************************************************************
 * MoioChart Library - graphs and charts made easy                         *
 * Copyright (C) 2005 by Silvio Moioli                                     *
 * silvio@moioli.net, http://www.moioli.net                                *  
 *                                                                         *
 * This program is free software; you can redistribute it and/or           *
 * modify it under the terms of the GNU General Public License             *
 * as published by the Free Software Foundation; either version 2          *
 * of the License, or (at your option) any later version.                  *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              *
 * MA  02110-1301, USA.                                                    *
 ***************************************************************************/

package net.moioli.chart;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Collects some of the necessary data for the DrawableI objects to be drawn on
 * a DefaultChart, and provides common methods to calculate some scale parameters.
 * 
 * @author Silvio Moioli
 * @version 2005/10/5
 */
public class Scale {
    // Constants
    /**
     * Since some mathematical functions may return the Infinity value under
     * certain conditions, we want a maximum y value for our DefaultChart. Values
     * beyond this constant (+/CUTVALUE) will never be drawn.
     */
    private static final double CUTVALUE = 100000;

    /**
     * Details (as little lines in the Legend or the Axes arrows) are
     * dimensioned with the formula (screenMaxX+screenMaxY)/(2*DETAILS).
     */
    private final static double DETAILS = 50;

    // Constructor-modified attributes
    /** Minimum represented abscissa. */
    private double minX = -5;

    /** Maximum represented abscissa. */
    private double maxX = 5;

    /** Minimum represented ordinate. */
    private double minY = -5;

    /** Maximum represented ordinate. */
    private double maxY = 5;

    /** Number of horizontal available pixels. */
    private int screenMaxX = 100;

    /** Number of vertical available pixels. */
    private int screenMaxY = 100;

    /**
     * To draw mathematical functions, DefaultChart will calculate the value of the
     * function multiple times. This parameter assigns the the number of pixels
     * on the x axis after which the y value will be recalculated. 1 is the
     * minimum possible value (maximum precision), 3 is recommended in most
     * cases.
     */
    private int pixelAccuracy = 3;

    /*
     * These attributes depend from the previous ones, and are recalculated by
     * the recalc() method.
     */
    /** Total x range width. */
    private double intX = 10;

    /** Total y range width. */
    private double intY = 10;

    /** x zoom factor. */
    private double xZoom = 10;

    /** y zoom factor. */
    private double yZoom = 10;

    /** Real value that corresponds to one pixel on the x axis. */
    private double step = 3 / 10;

    /**
     * Dimensions of a detail, calculated with the formula:
     * (screenMaxX+screenMaxY)/(2*DETAILS).
     */
    private int detail = 2;

    /** Default constructor, not available. */
    private Scale() {
        super();
    }

    /**
     * Autoscaling constructor. This one will try to guess minY and maxY values
     * querying the DefaultChart's DrawableI objects, and will adjust all the other
     * parameters accordingly.
     * 
     * @param c
     *            the DrawableI Collection of objects to be drawn
     * @param clip
     *            the onscreen rectangle on which the DefaultChart will be drawn
     * @param minX
     *            minimum represented abscissa
     * @param maxX
     *            maximum represented abscissa
     * @param pixelAccuracy
     *            accuracy, see the DefaultChart object
     */
    Scale(Collection c, IntRectangle clip, double minX, double maxX,
            int pixelAccuracy) {
        super();
        
        // Copies the x values
        this.minX = minX;
        this.maxX = maxX;

        // Compute available pixels
        this.screenMaxX = clip.getWidth() - 1;
        this.screenMaxY = clip.getHeight() - 1;

        // Setto la precisione
        this.pixelAccuracy = pixelAccuracy;

        // Calculates minimum and maximum y values
        this.minY = Double.POSITIVE_INFINITY;
        this.maxY = Double.NEGATIVE_INFINITY;
        Iterator i = c.iterator();
        while (i.hasNext()) {
            DrawableI d = (DrawableI) i.next();
            setNewY(d.minY(minX, maxX));
            setNewY(d.maxY(minX, maxX));
        }
        if (this.minY == Double.POSITIVE_INFINITY) {
            // None of the objects needed min y scaling
            this.minY = -1;
        }
        if (this.maxY == Double.NEGATIVE_INFINITY) {
            // None of the objects needed min y scaling
            this.maxY = 1;
        }
        // Strange things may happen sometimes if min and
        // max are too close
        if (Math.abs(this.minY - this.maxY) < 0.001) {
            this.minY = -5;
            this.maxY = 5;
        }
        // In any case, leave a little more space
        this.minY += this.minY / 10;
        this.maxY += this.maxY / 10;

        // Calculates dependent attributes
        recalc();
    }

    /**
     * Standard constructor.
     * 
     * @param clip
     *            the onscreen rectangle on which the DefaultChart will be drawn
     * @param minX
     *            minimum represented abscissa
     * @param maxX
     *            maximum represented abscissa
     * @param minY
     *            minimum represented ordinate
     * @param maxY
     *            maximum represented ordinate
     * @param pixelAccuracy
     *            accuracy, see the DefaultChart object
     */
    Scale(IntRectangle clip, double minX, double maxX, double minY,
            double maxY, int pixelAccuracy) {
        this(new ArrayList(0), clip, minX, maxX, pixelAccuracy);

        // Simply copy the y values
        this.minY = minY;
        this.maxY = maxY;

        // Calculates dependent attributes
        recalc();
    }

    /**
     * Calculates the internal attributes. Used by the constructors.
     */
    private void recalc() {
        this.intX = this.maxX - this.minX;
        this.intY = this.maxY - this.minY;
        this.xZoom = this.screenMaxX / this.intX;
        this.yZoom = this.screenMaxY / this.intY;
        this.step = this.pixelAccuracy / this.xZoom;
        this.detail = (int) Math.round((this.screenMaxX + this.screenMaxY)
                / (2 * DETAILS));
    }

    /**
     * Ensures the given y coordinate will be on the DefaultChart (except if it is
     * beyond the CUTVALUE limit).
     * 
     * @param y
     *            ordinate
     */
    private void setNewY(double y) {
        if (y > this.maxY) {
            this.maxY = y;
        }
        if (y < this.minY) {
            this.minY = y;
        }

        if (this.minY < -CUTVALUE) {
            this.minY = -CUTVALUE;
        }
        if (this.maxY > CUTVALUE) {
            this.maxY = CUTVALUE;
        }
    }

    // Useful methods
    /**
     * Calculates the abscissa on the screen (in pixels) corresponding to the
     * real abscissa passed according to this Scale object.
     * 
     * @param x
     *            real abscissa
     * @return screen abscissa
     */
    public int getX(double x) {
        return (int) Math.round(this.xZoom * (x - this.minX));
    }

    /**
     * Calculates the ordinate on the screen (in pixels) corresponding to the
     * real ordinate passed according to this Scale object.
     * 
     * @param y
     *            real abscissa
     * @return screen ordinate
     */
    public int getY(double y) {
        return (int) Math.round(this.screenMaxY
                - (this.yZoom * (y - this.minY)));
    }

    /**
     * When drawing mathematical functions it is needed to recalculate the
     * function's value every "dx" increment from minX to maxX. This method
     * returns the otimal increment value given the the scale, screen size and
     * the pixelAccuracy parameter.
     * 
     * @return the optimal x increment value
     */
    public double getStep() {
        return this.step;
    }

    /**
     * Returns the x zoom factor
     * 
     * @return x zoom factor
     */
    public double getXZoom() {
        return this.xZoom;
    }

    /**
     * Returns the y zoom factor
     * 
     * @return y zoom factor
     */
    public double getYZoom() {
        return this.yZoom;
    }

    /**
     * Given a segment parallel to the x axis which length is l, this method
     * returns the corresponding length in pixels.
     * 
     * @param l
     *            real length of a segment
     * @return "pixel length" of a segment
     */
    public int getXMeasure(double l) {
        return (int) Math.round(l * this.xZoom);
    }

    /**
     * Given a segment parallel to the y axis which length is l, this method
     * returns the corresponding length in pixels.
     * 
     * @param l
     *            real length of a segment
     * @return "pixel length" of a segment
     */
    public int getYMeasure(double l) {
        return (int) Math.round(l * this.yZoom);
    }

    // Getters
    /**
     * @return Returns the maximum represented abscissa.
     */
    public double getMaxX() {
        return this.maxX;
    }

    /**
     * @return Returns the maximum represented ordinate.
     */
    public double getMaxY() {
        return this.maxY;
    }

    /**
     * @return Returns the minimum represented abscissa.
     */
    public double getMinX() {
        return this.minX;
    }

    /**
     * @return Returns the minimum represented ordinate.
     */
    public double getMinY() {
        return this.minY;
    }

    /**
     * @return Returns the pixelAccuracy parameter.
     */
    public int getPixelAccuracy() {
        return this.pixelAccuracy;
    }

    /**
     * @return Returns the number of horizontal available pixels.
     */
    public int getScreenMaxX() {
        return this.screenMaxX;
    }

    /**
     * @return Returns the number of vertical available pixels.
     */
    public int getScreenMaxY() {
        return this.screenMaxY;
    }

    /**
     * @return Returns the detail parameter.
     */
    public int getDetail() {
        return this.detail;
    }
}