/***************************************************************************
 * MoioChart Library - graphs and charts made easy                         *
 * Copyright (C) 2005 by Silvio Moioli                                     *
 * silvio@moioli.net, http://www.moioli.net                                *  
 *                                                                         *
 * This program is free software; you can redistribute it and/or           *
 * modify it under the terms of the GNU General Public License             *
 * as published by the Free Software Foundation; either version 2          *
 * of the License, or (at your option) any later version.                  *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,              *
 * MA  02110-1301, USA.                                                    *
 ***************************************************************************/

package net.moioli.drawables;

import java.lang.Double;

import net.moioli.chart.DrawableI;
import net.moioli.chart.GraphicsI;
import net.moioli.chart.Scale;

/**
 * A general succession implementation.
 * 
 * @author Silvio Moioli
 * @version 2005/10/5
 */
public abstract class AbstractDrawableSuccession implements SuccessionI,
        DrawableI {

    /**
     * Default constructor.
     */
    public AbstractDrawableSuccession() {
        super();
    }

    /**
     * Calculates the series n a given n index.
     * 
     * @param n
     *            the series n index
     * @return the series value
     * @throws ArithmeticException
     *             if math errors are encountered
     */
    public abstract double f(int n);

    /**
     * Returns this.f(x), or NaN if an exception occured while computing.
     * 
     * @param n
     *            the abscissa
     * @return {an} or NaN
     */
    public double safeF(int n) {
        double result;
        try {
            if (n < 0) {
                throw new ArithmeticException();
            }
            result = this.f(n);
        } catch (ArithmeticException e) {
            result = Double.NaN;
        }
        return result;
    }

    /**
     * Implements Drawable, making it possible to draw this object.
     * 
     * @param g
     *            the surface where this Axes will be drawn
     * @param s
     *            the DefaultChart's Scale object
     * @param c
     *            the color to be used for drawing
     */
    public void drawMe(GraphicsI g, Scale s, Object c) {
        double minX = s.getMinX();
        double maxX = s.getMaxX();
        int intMinX = (int) Math.floor(minX);
        int intMaxX = (int) Math.ceil(maxX);
        for (int j = intMinX; j < intMaxX; j += 1) {
            double valore = safeF(j);
            if (!Double.isNaN(valore)) {
                new BigPoint(j, valore, "").drawMe(g, s, c);
            }
        }
    }

    /**
     * @see net.moioli.chart.DrawableI#minY(double, double)
     */
    public double minY(double minX, double maxX) {
        // old algorithm
        double minY = Double.POSITIVE_INFINITY;
        int intMinX = (int) Math.floor(minX);
        int intMaxX = (int) Math.ceil(maxX);
        for (int j = intMinX; j < intMaxX; j += 1) {
            double valore = safeF(j);
            if ((Double.isNaN(valore) == false) && (valore < minY)) {
                minY = valore;
            }
        }
        return minY;
    }

    /**
     * @see net.moioli.chart.DrawableI#maxY(double, double)
     */
    public double maxY(double minX, double maxX) {
        // old algorithm
        double maxY = Double.NEGATIVE_INFINITY;
        int intMinX = (int) Math.floor(minX);
        int intMaxX = (int) Math.ceil(maxX);
        for (int j = intMinX; j < intMaxX; j += 1) {
            double valore = safeF(j);
            if ((Double.isNaN(valore) == false) && (valore > maxY)) {
                maxY = valore;
            }
        }
        return maxY;
    }
}